package org.speleologist.common;

/*
 * Surveyor - the Poor Man's Digital Survey Device
 * Copyright Chris Hunter 2012 
 * email chris_at_arnside[dot]net
 * 
 * This file is part of Surveyor.

    Surveyor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Surveyor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Surveyor.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.LinkedList;

public class LineTracer implements Runnable {

	public boolean runNow;
	public ArrayList<ArrayList<int[]>> polyLines;
	public ArrayList<ArrayList<int[]>> pathGroups = new ArrayList<ArrayList<int[]>>(); // holds
																						// longer
																						// lines
	public LinkedList<int[]> trail;

	private static final int X = 0;
	private static final int Y = 1;
	private static final int P = 2;

	// private BufferedImage biBW;
	private int[][] pixels;

	private ArrayList<int[]> curPolyLine;
	private HashMap<PixelKey, int[]> polyIndex; // as traced short, two pixel
												// lines

	private int nl[] = new int[3]; // next left pixel
	private int nr[] = new int[3]; // next right
	private int nal[] = new int[3]; // next ahead left
	private int nar[] = new int[3];
	private boolean nh = false;

	private boolean showTrail = true;

	/*
	 * public LineTracer(BufferedImage biBW, Component cmpt) { this.biBW = biBW;
	 * this.component = cmpt;
	 * 
	 * }
	 */

	public LineTracer(int[][] lum) {
		this.pixels = lum;
	}

	public void run() {

		try {
			while (true) {
				if (runNow)
					trace();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}

	}

	/**
	 * Trace the lines in the bit map. Based on the PoTrace idea of finding an
	 * edge (white on the left, black on the right and then following it
	 * through, but this traces either edge of the black and stores the centre
	 * point
	 */
	public void trace() {
		polyLines = new ArrayList<ArrayList<int[]>>();
		polyIndex = new HashMap<PixelKey, int[]>();

		trail = new LinkedList<int[]>();

		traceEdge(showTrail);

		computePaths();
		runNow = false;
	}

	private void traceEdge(boolean showTrace) {
		int prevPxl = 1;
		for (int y = 0; y < pixels[0].length - 1; y++) {
			for (int x = 0; x < pixels.length - 1; x++) {

				if (prevPxl == 1 && getBrightness(x, y) == 0) {
					// don't bother if we processed it before
					if (polyIndex.containsKey(new PixelKey(x, y)))
						continue;

					// found a white/black transition. Look around for the next
					// black pixel,
					// since we are working right and down the raster ignore
					// above and left
					if (getBrightness(x, y + 1) == 0
							|| getBrightness(x - 1, y + 1) == 0) {
						// down
						setPixel(nl, x, y);
						setPixel(nr, x - 1, y);
						setPixel(nal, x, y + 1);
						setPixel(nar, x - 1, y + 1);
						nh = false;
					} else if (getBrightness(x + 1, y) == 0
							|| getBrightness(x + 1, y + 1) == 0) {
						// ahead
						setPixel(nl, x, y);
						setPixel(nr, x, y + 1);
						setPixel(nal, x + 1, y);
						setPixel(nar, x + 1, y + 1);
						nh = true;
					}

					curPolyLine = new ArrayList<int[]>();
					polyLines.add(curPolyLine);
					if (!checkAndAdd())
						continue;

					// loop while we can follow an edge
					while (followEdge(nl.clone(), nr.clone(), nal.clone(),
							nar.clone(), nh)) {

						if (!checkAndAdd())
							break;
						// are we off the map?
						if (nl[X] < -1 || nl[Y] < -1 || nl[X] >= pixels.length
								|| nl[Y] >= pixels[0].length)
							break;

						if (showTrace)
							drawTrail();
					}

					// draw as we go
					// if((polyLines.size() % 100) == 0) {
					// component.repaint();
					// }

					// clearFill();
				}

				prevPxl = getBrightness(x, y);
			}
		}
	}

	/**
	 * this method did recurse which was neater but ran out of stack space on
	 * large rasters
	 * 
	 * @param l
	 * @param r
	 * @param al
	 * @param ar
	 * @param horizontal
	 * @return
	 */
	private boolean followEdge(int[] l, int[] r, int[] al, int[] ar,
			boolean horizontal) {

		// follow the edge (black/white transition),
		// left and right are taken IN THE DIRECTION of travel
		//
		// x1|x2
		// |
		// y1 R|L r = pixel at x1,y1 l = pixel at x2,y1
		// ------------|-----
		// y2 AR|AL ar = pixel at x1,y2 al = pixel at x2,y2
		// |
		// V
		//
		//
		// x1|x2
		// |
		// y1 l|al r = pixel at x1,y2 l = pixel at x1,y1
		// ------------|----------->
		// y2 r|ar ar = pixel at x2,y2 al = pixel at x2,yy1
		// |
		//
		//

		// System.out.println(l[X] + "," + l[Y] + " ");

		int pixels = 0;
		pixels = pixels << 1;
		pixels = pixels | l[P];
		pixels = pixels << 1;
		pixels = pixels | r[P];
		pixels = pixels << 1;
		pixels = pixels | al[P];
		pixels = pixels << 1;
		pixels = pixels | ar[P];

		switch (pixels) {

		case 5: // 0101b
			// straight ahead
			setPixel(nl, al[X], al[Y]);
			setPixel(nr, ar[X], ar[Y]);
			// nl = al;
			// nr = ar;
			if (!horizontal) {
				nal[X] = al[X];
				nar[X] = ar[X];
				nal[Y] = al[Y] + al[Y] - l[Y];
				nar[Y] = ar[Y] + ar[Y] - r[Y];
				nh = false;
			} else {
				nal[X] = al[X] + al[X] - l[X];
				nar[X] = ar[X] + ar[X] - r[X];
				nal[Y] = al[Y];
				nar[Y] = ar[Y];
				nh = true;
			}
			setPixel(nal);
			setPixel(nar);

			break;

		case 7: // 0111b
		case 9: // 1001b: // left diagonal
			// turn left
			setPixel(nl, l[X], l[Y]);
			setPixel(nr, al[X], al[Y]);
			if (!horizontal) {
				nal[X] = l[X] + l[X] - r[X];
				nar[X] = l[X] + l[X] - r[X];
				nal[Y] = nl[Y];
				nar[Y] = nr[Y];
				nh = true;
			} else {
				nal[X] = nl[X];
				nar[X] = nr[X];
				nal[Y] = l[Y] + l[Y] - r[Y];
				nar[Y] = l[Y] + l[Y] - r[Y];
				nh = false;
			}
			setPixel(nal);
			setPixel(nar);
			break;

		case 4: // 0100b:
		case 6: // 0110b right diagonal
			// turn right
			setPixel(nl, ar[X], ar[Y]);
			setPixel(nr, r[X], r[Y]);
			// nl = ar;
			// nr = r;
			if (!horizontal) {
				nal[X] = r[X] + r[X] - l[X];
				nar[X] = r[X] + r[X] - l[X];
				nal[Y] = ar[Y];
				nar[Y] = r[Y];
				nh = true;
			} else {
				nal[X] = ar[X];
				nar[X] = r[X];
				nal[Y] = r[Y] + r[Y] - l[Y];
				nar[Y] = r[Y] + r[Y] - l[Y];
				nh = false;
			}
			setPixel(nal);
			setPixel(nar);
			break;

		default:
			return false;

		}

		return true;
	}

	int spinCt = 0; // stop it from spinning around a single point

	/**
	 * add a point to the current polygon
	 * 
	 * @param p
	 * @return true if added, false if tracing should stop
	 */
	private boolean checkAndAdd() {

		int[] node = { nl[X], nl[Y] };

		if (curPolyLine.size() > 0) {

			if (curPolyLine.size() > 2
					&& Arrays.equals(curPolyLine.get(0), node)) {
				// back at the beginning, close and end
				addToPolygon(node);
				return false;
			}

			// don't add it if it is the same as the last point
			if (Arrays.equals(curPolyLine.get(curPolyLine.size() - 1), node)) {
				if (spinCt++ < 3) {
					return true;
				} else {
					return false;
				}
			}
			spinCt = 0;

			// check other polygons, if it crosses stop
			// if(polyIndex.containsKey(new PixelKey(p[X],p[Y]))) return false;

			if (polyIndex.containsKey(new PixelKey(node))) {
				return false;
			}

		}

		addToPolygon(node);

		return true;

	}

	/**
	 * clear areas inside polygons
	 */
	/*
	 * private void clearFill() {
	 * 
	 * if (curPolyLine.size() == 0) return; //if it's not closed, just clear the
	 * line if(!Arrays.equals( curPolyLine.get(0),
	 * curPolyLine.get(curPolyLine.size()-1))) { for(int[] p : curPolyLine) {
	 * biBW.setRGB(p[X], p[Y], Color.WHITE.getRGB()); //biSrc.setRGB(p[X], p[Y],
	 * Color.ORANGE.getRGB()); } }
	 * 
	 * else { //create a sorted list TreeMap<Integer,ArrayList<Integer>> yx =
	 * new TreeMap<Integer,ArrayList<Integer>>();
	 * 
	 * for(int i = 0; i<curPolyLine.size()-2; i++) { Integer ix =
	 * curPolyLine.get(i)[X]; Integer iy = curPolyLine.get(i)[Y];
	 * 
	 * ArrayList<Integer> xs = yx.get(iy); if(xs == null) { xs = new
	 * ArrayList<Integer>(); yx.put(iy, xs); } xs.add(ix); }
	 * 
	 * 
	 * for(Integer cy : yx.keySet() ) { ArrayList<Integer> xs = yx.get(cy);
	 * Collections.sort(xs); if( xs.size() > 1) { //clear between contiguous
	 * groups and even/odd pairs int p=0; while( p<xs.size()) { int q=p+1;
	 * while(q<xs.size() && (xs.get(q-1)+1)==xs.get(q)){ q++; } if(q>=xs.size())
	 * q = xs.size()-1;
	 * 
	 * //the polygon may have a hole in it //TODO a better approach to holes may
	 * be to check for a hole //and process it prior to clearing the fill
	 * pixels. boolean hole=false; for(int cx=xs.get(p); cx<=xs.get(q); cx++) {
	 * 
	 * if( getBrightness(cx, cy ) == getBrightness(cx+1, cy )) { if(!hole) {
	 * biBW.setRGB(cx, cy, Color.WHITE.getRGB()); //biSrc.setRGB(cx, cy,
	 * Color.ORANGE.getRGB()); } else { //skip the hole edge p += 2; }
	 * hole=false; } else { //in a hole hole = true; }
	 * 
	 * } p=q+1; } } } }//else a closed polygon
	 * 
	 * 
	 * }
	 */

	private void drawTrail() {
		// draw a trail so that we can see what is happening
		int[] curPoint = { nl[X], nl[Y], nr[X], nr[Y], nal[X], nal[Y], nar[X],
				nar[Y] };
		trail.add(curPoint);
		if (trail.size() > 16) {
			trail.removeFirst();
		}

		// throttle back to give it time to draw
		try {
			Thread.sleep(1);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		// component.repaint();

	}

	private int getBrightness(int x, int y) {

		return pixels[x][y];
		/*
		 * if(x<biBW.getWidth() && y<biBW.getHeight() && x>=0 && y>=0) {
		 * 
		 * Color pxClr = new Color(biBW.getRGB(x, y)); float[] hsb =
		 * Color.RGBtoHSB(pxClr.getRed(), pxClr.getGreen(), pxClr.getBlue(),
		 * null);
		 * 
		 * return Math.round( hsb[2] ); } else { return 1; }
		 */
	}

	private void addToPolygon(int[] node) {
		int[] idx = { polyLines.size() - 1, curPolyLine.size() };
		polyIndex.put(new PixelKey(node[X], node[Y]), idx);
		curPolyLine.add(node);

	}

	private void setPixel(int[] point) {
		if (point[X] >= 0 && point[Y] >= 0 && point[X] < pixels.length
				&& point[Y] < pixels[0].length) {
			point[P] = getBrightness(point[X], point[Y]);
		} else {
			point[P] = 1;
		}
	}

	private void setPixel(int[] p, int x, int y) {
		p[X] = x;
		p[Y] = y;
		if (x >= 0 && y >= 0 && x < pixels.length && y < pixels[0].length) {
			p[P] = getBrightness(x, y);
		} else {
			p[P] = 1;
		}
	}

	/**
	 * tracing gave us poly-lines made up of short (2 pixel long) lines this
	 * reduces them to longer straight lines where possible by looking for one
	 * pixel steps
	 */
	private void computePaths() {

		for (ArrayList<int[]> poly : this.polyLines) {
			// lines are anything within 1 pixel of a line
			// projected from the start line where we have not gone in
			// all 4 directions

			int startIdx = 0;
			int finIdx = 2;
			ArrayList<int[]> curPaths = new ArrayList<int[]>();
			int[] prevP = { 0, 0, 0, 0 };
			while (finIdx < poly.size()) {
				finIdx = findPath(poly, startIdx, finIdx, 0);
				if (finIdx > poly.size() - 1)
					finIdx = poly.size() - 1;
				int[] p = { poly.get(startIdx)[X], poly.get(startIdx)[Y],
						poly.get(finIdx)[X], poly.get(finIdx)[Y] };
				if (!isNear(p, prevP)) {
					curPaths = new ArrayList<int[]>();
					pathGroups.add(curPaths);
				}

				prevP = p;

				curPaths.add(p);
				// repaint();

				startIdx = finIdx;
				finIdx = startIdx + 2;
			}

			if (poly.size() - startIdx == 2) {
				int[] p = { poly.get(0)[X], poly.get(0)[Y], poly.get(1)[X],
						poly.get(1)[Y] };
				curPaths.add(p);
			}

			// this.component.repaint();

		}

		// save some memory
		// polyLines.clear();

		// repaint();
	}

	private boolean isNear(int[] a, int[] b) {
		float dx1 = a[0] - b[0];
		float dy1 = a[1] - b[1];
		float dx2 = a[2] - b[2];
		float dy2 = a[3] - b[3];
		if (Math.sqrt(dx1 * dx1 + dy1 * dy1) < 3)
			return true;
		if (Math.sqrt(dx2 * dx2 + dy2 * dy2) < 3)
			return true;

		return false;

	}

		
	/**
	 * from a list of points, find adjacent contiguous points and join them up
	 * 
	 * @param poly
	 * @param sIdx
	 * @param cIdx
	 * @return
	 */
	private int findPath(ArrayList<int[]> poly, 
			int sIdx, int cIdx, int depthCount) {
		
		//this has stack overflowed due to recursion
		//it is unlikely that a line will be more than 20 points 
		//long so bail out if it is
		if(depthCount>20 ) return cIdx;
		
		if(cIdx >= poly.size()) return cIdx;
		
		float x1 = poly.get(sIdx)[X];
		float y1 = poly.get(sIdx)[Y];
		float x2 = poly.get(cIdx-1)[X];
		float y2 = poly.get(cIdx-1)[Y];
		float x3 = poly.get(cIdx)[X];
		float y3 = poly.get(cIdx)[Y];
		
		//double slope = (y2-y1)/(x2-x1);
		
		int result = cIdx-1;
		
			if(x3-x1 == 0) {
				//vert
				if(Math.abs(x2-x1) == 0 )
					result = findPath(poly, sIdx, cIdx+1, depthCount+1);
			}
			else if(y3-y1 == 0) {
				//horiz
				if(Math.abs(y2-y1) == 0 )
					result = findPath(poly, sIdx, cIdx+1, depthCount+1);
			}
			else {			
				float fx2 = x1+(x3-x1)/2;
				float fy2 = y1+(y3-y1)/2;
				float m = (y3-y1)/(x3-x1);
				float c = y1-(m*x1);
				float px2 = (fy2-c)/m;
				float py2 = m*fx2 + c;
				
				if( Math.abs(px2 - x2) <= 0.1
					&& Math.abs(py2 - y2) <= 0.1) {
					
					result = findPath(poly, sIdx, cIdx+1, depthCount+1);
				}
			}
				
		return result;
	}

	public class PixelKey {
		int x;
		int y;

		PixelKey(int x, int y) {
			this.x = x;
			this.y = y;
		}

		PixelKey(int[] xy) {
			this.x = xy[0];
			this.y = xy[1];
		}

		public boolean equals(Object that) {
			return (that instanceof PixelKey)
					&& (this.x == ((PixelKey) that).x && this.y == ((PixelKey) that).y);

		}

		/**
		 * need hashCode to match if x==x1 && y==y1
		 */
		public int hashCode() {
			return Arrays.hashCode(new int[] { x, y });
		}

	}

}
