package org.speleologist.dangle;

/*
 * Surveyor - the Poor Man's Digital Survey Device
 * Copyright Chris Hunter 2012 
 * email chris_at_arnside[dot]net
 * 
 * This file is part of Surveyor.

    Surveyor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Surveyor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Surveyor.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.IOException;

import android.content.Context;
import android.hardware.Camera;
import android.hardware.Camera.PreviewCallback;
import android.util.Log;
import android.view.SurfaceHolder;
import android.view.SurfaceView;
import android.widget.Toast;

/** A basic Camera preview class */
public class CameraWrapper extends SurfaceView implements
		SurfaceHolder.Callback {
	private static final String TAG = "CameraPreview";
	private SurfaceHolder mHolder;
	private Camera mCamera;
	private boolean cameraActive;
	private PreviewCallback callback;
	private Context context;

	public CameraWrapper(Context context, PreviewCallback callback) {
		super(context);
		
		this.context = context;
		this.callback = callback;
		// Install a SurfaceHolder.Callback so we get notified when the
		// underlying surface is created and destroyed.
		mHolder = getHolder();
		mHolder.addCallback(this);
		// deprecated setting, but required on Android versions prior to 3.0
		mHolder.setType(SurfaceHolder.SURFACE_TYPE_PUSH_BUFFERS);
	}

	public void surfaceCreated(SurfaceHolder holder) {
		// The Surface has been created, now tell the camera where to draw the
		// preview.
		setupCamera();
		try {
			mCamera.setPreviewDisplay(holder);
			mCamera.startPreview();
		} catch (IOException e) {
			Log.d(TAG, "Error setting camera preview: " + e.getMessage());
		}
	}

	public void refreshSurface() {
		surfaceCreated(mHolder);
	}

	public void surfaceDestroyed(SurfaceHolder holder) {
		// empty. Take care of releasing the Camera preview in your activity.
		shutDownCamera();
	}

	public void surfaceChanged(SurfaceHolder holder, int format, int w, int h) {
		// If your preview can change or rotate, take care of those events here.
		// Make sure to stop the preview before resizing or reformatting it.

		if (mHolder.getSurface() == null) {
			// preview surface does not exist
			return;
		}

		// stop preview before making changes
		try {
			mCamera.stopPreview();
		} catch (Exception e) {
			// ignore: tried to stop a non-existent preview
		}

		// set preview size and make any resize, rotate or
		// start preview with new settings
		try {
			if (cameraActive) {
				setupCamera();
				mCamera.setPreviewDisplay(mHolder);
				mCamera.startPreview();
			}

		} catch (Exception e) {
			Log.d(TAG, "Error starting camera preview: " + e.getMessage());
		}
	}

	/**
	 * Get the camera object
	 */
	public Camera setupCamera() {
		if (mCamera == null) {
			try {
				mCamera = Camera.open(); // attempt to get a Camera instance
				
				Camera.Parameters p = mCamera.getParameters();
				p.setRotation(90);
				mCamera.setParameters(p);

				mCamera.setDisplayOrientation(90);
				
				mCamera.setPreviewCallback(callback);
			} catch (Exception e) {
				Log.e(TAG, "Execption setting up camera", e);
				Toast.makeText(context, e.getLocalizedMessage(), Toast.LENGTH_LONG);
			}
		}
		return mCamera; // returns null if camera is unavailable
	}

	public void shutDownCamera() {
		if(mCamera!=null){
			mCamera.stopPreview();
			mCamera.setPreviewCallback(null);
			mCamera.release();	
			mCamera=null;
			}
	}

	public boolean isCameraActive() {
		return cameraActive;
	}

	public void setCameraActive(boolean previewActive) {
		this.cameraActive = previewActive;
		if (previewActive) {
			setupCamera();
			mCamera.startPreview();
		} else {
			shutDownCamera();
		}
	}
}