package org.speleologist.dangle;

/*
 * Surveyor - the Poor Man's Digital Survey Device
 * Copyright Chris Hunter 2012 
 * email chris_at_arnside[dot]net
 * 
 * This file is part of Surveyor.

    Surveyor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Surveyor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Surveyor.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.util.ArrayList;

import android.content.Context;
import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Matrix;
import android.graphics.Paint;
import android.graphics.Path;
import android.graphics.Rect;
import android.graphics.RectF;
import android.util.Log;
import android.view.MotionEvent;
import android.view.SurfaceView;
import android.view.View;
import android.view.animation.Animation;
import android.view.animation.LinearInterpolator;
import android.view.animation.RotateAnimation;
import android.widget.ImageView;
import android.widget.TextView;
import android.widget.Toast;

public class DrawView extends ImageView {
	
	private int[][] pixels;
	private int[][] digitArea;
	private int[][] procPix;
	private int width ;
	private int height;
	
	private ArrayList<ArrayList<int[]>> tracedLines;
	private RectF rect;
	private boolean areaSelectable;
	private TextView txtCoord;

	public DrawView(Context context, int[][] pixels, int[][]digitArea,
			TextView txtCoord) {
		super(context);

		this.pixels = pixels;
		this.digitArea = digitArea;
		this.txtCoord = txtCoord;

	    width =  pixels.length;
		height = pixels[0].length;
	}

	@Override
	public void onDraw(Canvas canvas) {

		//TODO this is just random it needs sorting out
		//and it needs changing in transformRect()
		canvas.rotate(90, width/2, 20+height/6);
		
		drawBitmap(canvas);

		if (procPix != null) {
			drawProcessedPixels(canvas);
		}

		if (tracedLines != null) {
			drawTrace(canvas);
		}
				
		
		if(rect!=null) {
			Paint rectP = new Paint();
			rectP.setAntiAlias(true);
			rectP.setStrokeWidth(2f);
			rectP.setColor(Color.GREEN);
			rectP.setStyle(Paint.Style.STROKE);
			rectP.setStrokeJoin(Paint.Join.ROUND);
			
			RectF rx = transformRect();
			canvas.drawRect(rx, rectP);
			//Log.i( "DrawView", String.format("Rect: %f, %f, %f, %f", rect.left, rect.top, rect.right, rect.bottom));
			//Log.i( "DrawView", String.format("Rect: %f, %f, %f, %f", rect.left-height/2, rect.top-height/2, rect.right-height/2, rect.bottom-height/2));
			//Log.i( "DrawView", String.format("RX: %f, %f, %f, %f", rx.left, rx.top, rx.right, rx.bottom));
		}

	}
	
	
	private RectF transformRect() {
		Matrix xform = new Matrix();
		xform.preRotate(270,  width/2, 20+height/6);
		RectF rx = new RectF();
		xform.mapRect(rx, rect);
		
		return rx;
	}
	
	 @Override
	  public boolean onTouchEvent(MotionEvent event) {
		 if(areaSelectable) {
			float eventX = event.getX();
			float eventY = event.getY();
			
			switch (event.getAction()) {
			case MotionEvent.ACTION_DOWN:
			  //path.moveTo(eventX, eventY);
			  rect = new RectF(eventX, eventY, eventX, eventY);
			  return true;
			case MotionEvent.ACTION_MOVE:
			  //path.lineTo(eventX, eventY);
				rect = new RectF(rect.left, rect.top, eventX, eventY);
				// Schedules a repaint.
				invalidate();
			  break;
			case MotionEvent.ACTION_UP:
				RectF rx = transformRect();
				//TODO not great doing this a Toast but better than nothing
				txtCoord.setVisibility(VISIBLE);
				txtCoord.setText(String.format("%3.0f, %3.0f, %3.0f, %3.0f", 
						rx.left, rx.top, rx.right, rx.bottom));
				Log.i( "DrawView", String.format("RX: %f, %f, %f, %f", rx.left, rx.top, rx.right, rx.bottom));
				
			  break;
			default:
			  return false;
			}
			
			
		 }
	    return true;
	  }

		 
	private void drawBitmap(Canvas canvas) {

		Bitmap image = Bitmap.createBitmap( width, height, Bitmap.Config.ARGB_8888);

		for (int y = 0; y < height; y++) {
			for (int x = 0; x < width; x++) {
				int newPixel = Color.argb(255, pixels[x][y], pixels[x][y], pixels[x][y]);
				image.setPixel(x, y, newPixel);
			}
		}
				
		canvas.drawBitmap(image, 0, 0, new Paint());
		
		Paint rectP = new Paint();
		rectP.setAntiAlias(true);
		rectP.setStrokeWidth(2f);
		rectP.setColor(Color.RED);
		rectP.setStyle(Paint.Style.STROKE);
		rectP.setStrokeJoin(Paint.Join.ROUND);
		canvas.drawRect(new Rect(digitArea[0][0], digitArea[0][1], digitArea[1][0], digitArea[1][1]), rectP);		
		
	}

	private void drawProcessedPixels(Canvas canvas) {

		Paint paint = new Paint();
		paint.setARGB(100, 200, 200, 90);
		for (int x = 0; x < procPix.length; x++) {
			for (int y = 0; y < procPix[0].length; y++) {
				Rect r = new Rect(x+100, y, x + 101, y+1);
				if (procPix[x][y] == 0)
					canvas.drawRect(r, paint);
			}
		}
	}

	private void drawTrace(Canvas canvas) {
		Paint paint = new Paint();
		for (int gr = 0; gr < tracedLines.size(); gr++) {
			// Alternate colours
			if (gr % 2 > 0)
				paint.setARGB(200, 0, 255, 0);
			else
				paint.setARGB(200, 0, 0, 255);

			// paths contains grouped paths with coincident ends
			ArrayList<int[]> paths = tracedLines.get(gr);
			for (int i = 0; i < paths.size(); i++) {
				int[] path = paths.get(i);
				if (path != null) {
					canvas.drawLine(path[0]+100, path[1], path[2]+100, path[3], paint);
				}
			}
		}

	}

	public void setProcessedPixels(int[][] pixels) {
		this.procPix = pixels;
	}

	public void setTracedLines(ArrayList<ArrayList<int[]>> tracedLines) {
		this.tracedLines = tracedLines;
	}

	public boolean isAreaSelectable() {
		return areaSelectable;
	}

	public void setAreaSelectable(boolean areaSelectable) {
		this.areaSelectable = areaSelectable;
		if(areaSelectable) {
			txtCoord.setVisibility(VISIBLE);
		}
		else {
			txtCoord.setVisibility(INVISIBLE);
			rect = null;
		}
	}

	public void setDigitArea(int[][] digitArea) {
		this.digitArea = digitArea;
		rect = null;
		invalidate();
	}

}
