package org.speleologist.dangle;

/*
 * Surveyor - the Poor Man's Digital Survey Device
 * Copyright Chris Hunter 2012 
 * email chris_at_arnside[dot]net
 * 
 * This file is part of Surveyor.

    Surveyor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Surveyor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Surveyor.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;

import org.speleologist.common.LcdDigitOcr;
import org.speleologist.common.Orientation;
import org.speleologist.common.OrientationListener;

import android.app.Activity;
import android.app.AlertDialog;
import android.app.Dialog;
import android.content.Context;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.SharedPreferences;
import android.content.pm.ActivityInfo;
import android.hardware.Camera;
import android.hardware.Camera.PreviewCallback;
import android.hardware.Camera.Size;
import android.hardware.Sensor;
import android.hardware.SensorManager;
import android.os.Bundle;
import android.os.Environment;
import android.os.Handler;
import android.preference.PreferenceManager;
import android.provider.Settings;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.MotionEvent;
import android.view.View;
import android.view.View.OnTouchListener;
import android.view.ViewGroup;
import android.widget.Button;
import android.widget.CheckBox;
import android.widget.EditText;
import android.widget.FrameLayout;
import android.widget.LinearLayout;
import android.widget.TextView;
import android.widget.Toast;
import android.widget.ViewFlipper;

public class SurveyorActivity extends Activity 
implements PreviewCallback,	OrientationListener, OnTouchListener  {

	private boolean bCapture;
	private CheckBox chkSection;

	private int imgWidth;
	private int imgHeight;
	private SharedPreferences sharedPrefs;

	private LcdDigitOcr ocr;
	private Orientation orient;

	private CameraWrapper mPreview;
	private FrameLayout previewFrame;
	private LinearLayout rootLayout;
	private SensorManager m_sensorManager;
	private AlertDialog alertDialog;
	private int outsetFlightMode;
	private boolean orientLocked;
	private String fromStn;
	private String toStn;
	
	private String folderName;
	private boolean bearingTouch;

	/** Called when the activity is first created. */
	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);

		outsetFlightMode = Settings.System.getInt(
			      getContentResolver(), 
			      Settings.System.AIRPLANE_MODE_ON, 0);		
		
		setFlightMode(1);		
		
		orient = new Orientation(this);

		bCapture = false;
		orientLocked = false;
		bearingTouch = false;
		
		sharedPrefs = PreferenceManager.getDefaultSharedPreferences(this);
		folderName = sharedPrefs.getString("data_folder", "Surveyor");

		setContentView(R.layout.main);
		setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_PORTRAIT);

		m_sensorManager = (SensorManager) getSystemService(Context.SENSOR_SERVICE);
		registerListeners();

		// Create our Preview view and set it as the content of our activity.
		mPreview = new CameraWrapper(this, this);
		mPreview.setCameraActive(false);
		previewFrame = (FrameLayout) this.findViewById(R.id.camera_preview);
		previewFrame.addView(mPreview);
		
		rootLayout = (LinearLayout) this.findViewById(R.id.rootLayout);
		
		rootLayout.setOnTouchListener(this);
	}

	
	private void registerListeners() {
		// m_sensorManager.registerListener(orient,
		// m_sensorManager.getDefaultSensor(Sensor.TYPE_ORIENTATION),
		// SensorManager.SENSOR_DELAY_GAME);

		m_sensorManager.registerListener(orient,
				m_sensorManager.getDefaultSensor(Sensor.TYPE_MAGNETIC_FIELD),
				SensorManager.SENSOR_DELAY_NORMAL);
		m_sensorManager.registerListener(orient,
				m_sensorManager.getDefaultSensor(Sensor.TYPE_ACCELEROMETER),
				SensorManager.SENSOR_DELAY_NORMAL);
	}

	private void unregisterListeners() {
		m_sensorManager.unregisterListener(orient);
	}

	@Override
	public void onDestroy() {
		unregisterListeners();
		mPreview.setCameraActive(false);
		mPreview = null;
		m_sensorManager = null;
		ocr = null;
		orient = null;
		
		setFlightMode(0);

		super.onDestroy();
	}

	@Override
	public void onPause() {
		unregisterListeners();
		setFlightMode(0);
		
		super.onPause();
		mPreview.setCameraActive(false);
		
		
	}

	@Override
	public void onResume() {
		registerListeners();
		super.onResume();
		
		setFlightMode(1);
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {

		// Inflate our menu which can gather user input for switching camera
		MenuInflater inflater = getMenuInflater();
		inflater.inflate(R.menu.menu, menu);
		return true;
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		if (item.getItemId() == R.id.settings) {
			startActivity(new Intent(this, Preferences.class));
		}
		else if (item.getItemId() == R.id.viewer) {
			startActivity(new Intent(this, YuvViewerActivity.class));
		}

		return true;
	}
	
	
	public boolean onTouch(View v, MotionEvent ev) {
		int action = ev.getAction() & MotionEvent.ACTION_MASK;
		
		if(bearingTouch) {
			bearingTouch = false;		    
		   
	    	//lock the orientation and read the Laser LCD 
			this.orientLocked = true;		
			
			Handler handler = new Handler();  				  		
			showPreview();
			handler.postDelayed(new Runnable() { 
	  	         public void run() { 
	  	        	captureDistance(); 
	  	         } 
	  	    }, 3000);
	   
		}
		
		/*if (action == MotionEvent.ACTION_DOWN) {
		        // Play "click" sound
	    } else if (action == MotionEvent.ACTION_UP || action == MotionEvent.ACTION_CANCEL) {
	    	
	    }*/
	    
	    return true;
	}

	

	private void setFlightMode(int state) {
		
		int curVal = Settings.System.getInt(
			      getContentResolver(), 
			      Settings.System.AIRPLANE_MODE_ON, 0);
		if(state != curVal) {
			Settings.System.putInt(super.getContentResolver(),
					Settings.System.AIRPLANE_MODE_ON, state);
			
			Intent intent = new Intent(Intent.ACTION_AIRPLANE_MODE_CHANGED);
			intent.putExtra("state", state==1);
			sendBroadcast(intent);
		}
	}
	
	
	public void orientationChanged() {
		if(!orientLocked) {
			orient.setSwapPitchRollLandscape(this.sharedPrefs.getBoolean(
					"swap_landscape", false));
			orient.computeOrientation(getResources().getConfiguration().orientation);
			TextView brg = (TextView) findViewById(R.id.bearing);
			brg.setText(String.format("%05.1f", orient.getYaw()));
			TextView clino = (TextView) findViewById(R.id.clino);
			clino.setText(String.format("%05.1f", orient.getPitch()));
		}

	}

	
	/**
	 * the Leg button has been clicked
	 * 
	 * Expects the laser to be on (user responsibility), sets up stations,
	 * blanks the screen and reads the bearing,
	 * waits for the user to read the distance then grabs the camera image and
	 * OCRs it.
	 * 
	 * This is fairly complex, but Android seems to like call backs
	 */
	public void legClick(View arg0) {
				
		final AlertDialog stnDialog = new AlertDialog.Builder(this).create();
		LayoutInflater inflater = getLayoutInflater();
		final View stnView = inflater.inflate(R.layout.stations, rootLayout, false);
		stnDialog.setView(stnView);
		
		if(toStn!=null) {
			EditText txtFromStn = (EditText) stnView.findViewById(R.id.txtFromStn);
			txtFromStn.setText(toStn);
		}
		
		CaptureHandler listener = new CaptureHandler();
		listener.txtFromStn = (EditText) stnView.findViewById(R.id.txtFromStn);
		listener.txtToStn = (EditText) stnView.findViewById(R.id.txtToStn);
		
		stnDialog.setButton("OK to Read Brg", listener);
		
		stnDialog.setButton2("Cancel", new DialogInterface.OnClickListener() {		
			public void onClick(DialogInterface dialog, int which) {
			
			}
		});
		stnDialog.show();
	}
	
	
	/**
	 * the Sect button has been clicked
	 * 
	 * Expects the laser to be on (user responsibility), sets up stations,
	 * blanks the screen and reads the bearing,
	 * waits for the user to read the distance then grabs the camera image and
	 * OCRs it.
	 * 
	 * This is fairly complex, but Android seems to like call backs
	 */
	public void sectClick(View view) {
				
		chkSection = (CheckBox) view;
		if(chkSection.isChecked()) {
			final AlertDialog stnDialog = new AlertDialog.Builder(this).create();
			LayoutInflater inflater = getLayoutInflater();
			final View stnView = inflater.inflate(R.layout.section, rootLayout, false);
			stnDialog.setView(stnView);
			
			if(toStn!=null) {
				EditText txtStn = (EditText) stnView.findViewById(R.id.txtBsStn);
				txtStn.setText(toStn);
			}
			
			CaptureHandler listener = new CaptureHandler();
			listener.txtFromStn = (EditText) stnView.findViewById(R.id.txtBsStn);
			listener.txtToStn = new EditText(super.getApplicationContext());
			listener.txtToStn.setText("Sect");
			stnDialog.setButton("OK", listener); 
			stnDialog.setButton2("Cancel", new DialogInterface.OnClickListener() {		
				public void onClick(DialogInterface dialog, int which) {
					chkSection.setChecked(false);
				}
			});
			stnDialog.show();
		}
		else {
			chkSection.setChecked(false);
		}
	}
	
	
	private void captureSection() {
		alertDialog = new AlertDialog.Builder(this).create();
		alertDialog.setMessage("OK to read offset,\nCancel to exit Section mode");
		
		CaptureHandler listener = new CaptureHandler();
		listener.txtFromStn = new EditText(super.getApplicationContext());
		listener.txtFromStn.setText(fromStn);
		listener.txtToStn = new EditText(super.getApplicationContext());
		listener.txtToStn.setText("Sect");
		alertDialog.setButton("OK", listener); 
		
		alertDialog.setButton2("Cancel", new DialogInterface.OnClickListener() {		
			public void onClick(DialogInterface dialog, int which) {
				chkSection.setChecked(false);
			}
		});
		alertDialog.show();
	}
	
	
	/**
	 * capture the current bearing
	 */
	private void captureOrient() {
		Toast toast = Toast.makeText(getApplicationContext(), 
				"Calibrate and point, touch screen when ready", Toast.LENGTH_SHORT);
		toast.show();
		
		bearingTouch = true;
		 
	}
	
	
	/**
	 * capture the Laser LCD and OCR it
	 */
	private void captureDistance() {
		alertDialog = new AlertDialog.Builder(this).create();
		alertDialog.setMessage("Measure distance\nOK when done");
		alertDialog.setButton("OK", new DialogInterface.OnClickListener() {
		      public void onClick(DialogInterface dialog, int which) {
		  		bCapture = true;
		  		
		  		//TODO TESTING ONLY
		  		//previewTest();		  		
		  		//alertDialog.dismiss();
		    } }); 
		alertDialog.show();
	}

	
	/**
	 * load the OCR 
	 */
	private void loadOcr() {
		File root = new File(Environment.getExternalStorageDirectory(),
				folderName);
		if (!root.exists()) {
			root.mkdirs();
		}

		//expecting "x1,y1,x2,y2"
		String imgArea = sharedPrefs.getString("image_area", "Surveyor");
		String[] parts = imgArea.split("[,. ]+");
		
		int x1=Integer.parseInt(parts[0]);
		int y1=Integer.parseInt(parts[1]);
		int x2=Integer.parseInt(parts[2]);
		int y2=Integer.parseInt(parts[3]);
		
		String sThreshold = sharedPrefs.getString("bw_threshold", "Surveyor");
		int threshold = Integer.parseInt(sThreshold);

		ocr = new LcdDigitOcr(new int[][]{{x1,y1},{x2,y2}}, threshold);
		
	}

	
	
	
	/**
	 * increment the number at the end of the station field
	 * @param arg0
	 */
	public void nextStationClick(View view) {
		EditText txt = (EditText) ((LinearLayout)view.getParent()).findViewById(R.id.txtToStn);
		String stn = txt.getText().toString();
		
		int num = 0;
		if(stn==null || stn.length()==0){
			txt.setText(String.valueOf(++num));
		}
		else {			
			int p = stn.length();
			while(p>0 && Character.isDigit(stn.charAt(p-1)) ) p--;
			
			if(p<stn.length()) num = Integer.parseInt(stn.substring(p));
			
			txt.setText(stn.substring(0,p) + String.valueOf(++num));
		}		
	}

	
	/**
	 * toggle the preview frame on and off
	 * @param arg0
	 */
	public void previewClick(View arg0) {
		mPreview.setCameraActive(!mPreview.isCameraActive());
		if (mPreview.isCameraActive()) {
			showPreview();
		} else {
			hidePreview();
		}
	}
	
	
	private void showPreview() {
		previewFrame.setVisibility(View.VISIBLE);
		mPreview.refreshSurface();
		rootLayout.invalidate();
	}
	
	
	private void hidePreview(){
		previewFrame.setVisibility(View.INVISIBLE);
		rootLayout.invalidate();
	}
	

	/**
	 * if ready to read the Laser LCD, grab the preview and process it
	 */
	public void onPreviewFrame(byte[] data, Camera camera) {
		if (bCapture) {
			bCapture = false;

			Camera.Parameters parameters = camera.getParameters();
			Size previewSize = parameters.getPreviewSize();

			if (ocr == null) {
				imgWidth = previewSize.width;
				imgHeight = previewSize.height;
				loadOcr();
			}
			
			boolean save = this.sharedPrefs.getBoolean("save_images", false);	
			int[] lum = new int[data.length];
			decodeYUV420SP(lum, data, imgWidth, imgHeight);
			int[][]pixels = new int[imgWidth][imgHeight];
			int lCt=0;
			for(int y=0; y<imgHeight; y++) {
				for (int x = 0; x < imgWidth; x++) {
					int c = lum[lCt++];
					pixels[x][y] = c;					
				}
			}
			
			if(save) saveImage(pixels);
			
			String dist = ocr.readValue(pixels);
			TextView tvd = (TextView) this.findViewById(R.id.dist);
			if(!dist.contains("_")) {
				float d = Float.parseFloat(dist)/1000;					
				tvd.setText(String.format("%02.3f", d));
			}
			else{
				tvd.setText(dist);
			}
			
			saveData();
			alertDialog.dismiss();
			this.orientLocked = false;	
			
			if(chkSection != null && chkSection.isChecked()) {
				Handler handler = new Handler(); 
		  	    handler.postDelayed(new Runnable() { 
		  	         public void run() { 
						captureSection();
		  	         } 
		  	    }, 2000); 
			}
		}		
	}
	
	
	/**
	 * just a test harness to test preview in emulator
	 */
	private void previewTest() {
		if (ocr == null) {
			imgWidth = 320;
			imgHeight = 240;
			loadOcr();
		}
		
		boolean save = this.sharedPrefs.getBoolean("save_images", false);				
		try {
			File file = new File(Environment.getExternalStorageDirectory()
					+ "/" + folderName, "out001.lum");
			DataInputStream dis = new DataInputStream(new FileInputStream(file));
			
			int[][]lum = new int[imgWidth][imgHeight];
			for(int y=0; y<imgHeight; y++) {
				for (int x = 0; x < imgWidth; x++) {
					int c = dis.readInt();
					lum[x][y] = c;					
				}
			}
			
			if(save) saveImage(lum);
			
			dis.close();
			String dist = ocr.readValue(lum);
			TextView tvd = (TextView) this.findViewById(R.id.dist);
			if(!dist.contains("_")) {
				float d = Float.parseFloat(dist)/1000;					
				tvd.setText(String.format("%06.3f", d));
			}
			else{
				tvd.setText(dist);
			}				
		}
		catch(Exception e) {
			Log.e("Surveyor", "Error reading test data", e);
		}
	}


	/**
	 * Extract the luminance (Y) values from a YUV image
	 * @param rgb
	 * @param yuv420sp
	 * @param width
	 * @param height
	 */
	static public void decodeYUV420SP(int[] lum, byte[] yuv420sp, int width,
			int height) {
		final int frameSize = width * height;

		for (int j = 0, yp = 0; j < height; j++) {
			int uvp = frameSize + (j >> 1) * width, u = 0, v = 0;
			for (int i = 0; i < width; i++, yp++) {
				int y = (0xff & ((int) yuv420sp[yp])) - 16;
				if (y < 0) y = 0;
				
				lum[yp] = y; 				
			}
		}
	}	

	
	/**
	 * save current values to a file
	 */
	private void saveData() {
		
		File file = new File(Environment.getExternalStorageDirectory() + 
				"/"+ folderName, this.sharedPrefs.getString("survey_file", "Survey.txt"));
		try {			
			PrintWriter  out = new PrintWriter (new FileWriter(file, true));
			TextView dist = (TextView) this.findViewById(R.id.dist);
			TextView brg = (TextView) findViewById(R.id.bearing);
			TextView clino = (TextView) findViewById(R.id.clino);
			out.format("%s - %s: %s, %s, %s%n", this.fromStn, this.toStn,
					brg.getText(), clino.getText(), dist.getText());
			out.flush();
			out.close();
		} catch (FileNotFoundException e) {
			Log.e("Surveyor", "Error saving data", e);
			Toast toast = Toast.makeText(
					this,
					"Could not save data, FileNotFound Exception\n"
							+ file.getAbsolutePath(), Toast.LENGTH_LONG);
			toast.show();
		} catch (IOException e) {
			Log.e("Surveyor", "Error saving data", e);
			Toast toast = Toast.makeText(
					this,
					"Could not save data, IOException Exception\n"
							+ file.getAbsolutePath(), Toast.LENGTH_LONG);
			toast.show();
		}
	}
	
	
	/**
	 * save the image data if OCR fails
	 * @param pixels
	 */
	private void saveImage(int[][] pixels) {
		try{
			File file = getNextFile(fromStn + "-" + toStn + "-img", "lum");
			DataOutputStream out = new DataOutputStream(
					new FileOutputStream(file));
			for(int y=0; y<imgHeight; y++) {
				for (int x = 0; x < imgWidth; x++) {
					out.writeByte(pixels[x][y]);
				}
			}
			out.close();
		}
					
		catch (FileNotFoundException e) {
			e.printStackTrace();
		} 
		catch (IOException e) {
			e.printStackTrace();
		}
			
	}
	
	
	/**
	 * get the name for a new file by appending the next available number to
	 * a root name
	 * 
	 * @param nameroot
	 * @param ext
	 * @return
	 */
	private File getNextFile(String nameroot, String ext) {
		File root = new File(Environment.getExternalStorageDirectory(),
				folderName);
		if (!root.exists()) {
			root.mkdirs();
		}

		int fCt = 1;
		File file = new File(root, String.format("%s%03d.%s", nameroot, fCt,
				ext));
		while (file.exists()) {
			file = new File(root,
					String.format("%s%03d.%s", nameroot, fCt++, ext));
		}

		return file;
	}

	
	
	/**
	 * Handle the Station dialog on Click
	 *
	 */
	private class CaptureHandler implements DialogInterface.OnClickListener {
				
		EditText txtFromStn;
  		EditText txtToStn; 

  		/** 
  		 * store the station data and prep for reading the bearing
  		 */
		public void onClick(DialogInterface dialog, int which) {
			
	  		fromStn =  txtFromStn.getText().toString();
	  		toStn = txtToStn.getText().toString();
			
			((Dialog) dialog).hide();
	  		
	  		//clear the screen and read the orientation 
	  		//(bright screen interferes with the magnetometers)
			hidePreview();
	  		//wait for screen to clear before reading bearing
	  		Handler handler = new Handler(); 
	  	    handler.postDelayed(new Runnable() { 
	  	         public void run() { 
	  	        	captureOrient(); 
	  	         } 
	  	    }, 2000); 		  
			
		}
	}

	
}