package org.speleologist.dangle;

/*
 * Surveyor - the Poor Man's Digital Survey Device
 * Copyright Chris Hunter 2012 
 * email chris_at_arnside[dot]net
 * 
 * This file is part of Surveyor.

    Surveyor is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Surveyor is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Surveyor.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FilenameFilter;
import java.io.IOException;

import org.speleologist.common.LcdDigitOcr;

import android.app.Activity;
import android.app.AlertDialog;
import android.app.AlertDialog.Builder;
import android.app.Dialog;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.SharedPreferences;
import android.content.pm.ActivityInfo;
import android.graphics.Color;
import android.os.Bundle;
import android.os.Environment;
import android.preference.PreferenceManager;
import android.util.Log;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.widget.FrameLayout;
import android.widget.TextView;
import android.widget.Toast;

public class YuvViewerActivity extends Activity 
implements DialogInterface.OnClickListener {
	
	DrawView drawView;
	private String[] mFileList;
    private File mPath;
    private String mChosenFile;
    private static final String FTYPE = ".lum";    
    private static final int DIALOG_LOAD_FILE = 1000;
	private static final String TAG = "YuvViewer";
	private SharedPreferences sharedPrefs;
	
	private int imgHeight;
	private int imgWidth;
	private int[][]pixels;
	private int[][] digitArea;

	private FrameLayout imageLayout;
		
    /** Called when the activity is first created. */
    @Override
    public void onCreate(Bundle savedInstanceState) {
    	 super.onCreate(savedInstanceState);
    	 
    	 mPath = new File(Environment.getExternalStorageDirectory() + "/Surveyor/");        

    	 setContentView(R.layout.viewer);
 		 setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_PORTRAIT);
    	 
    	 imageLayout = (FrameLayout) findViewById(R.id.viewer_image);
    	 sharedPrefs = PreferenceManager.getDefaultSharedPreferences(this);    	 
    	 loadDigitArea();
    }
    
    
    private void loadDigitArea() {
    	String imgArea = sharedPrefs.getString("image_area", "0,0,320,240");
 		String[] parts = imgArea.split("[, ]+");
 		
 		int x1=Integer.parseInt(parts[0]);
 		int y1=Integer.parseInt(parts[1]);
 		int x2=Integer.parseInt(parts[2]);
 		int y2=Integer.parseInt(parts[3]);
 		
 		digitArea = new int[][]{{x1,y1},{x2,y2}};
    }
    
    
    @Override
	public boolean onCreateOptionsMenu(Menu menu) {

		// Inflate our menu which can gather user input 
		MenuInflater inflater = getMenuInflater();
		inflater.inflate(R.menu.viewermenu, menu);
		return true;
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		if (item.getItemId() == R.id.open) {
			showDialog(DIALOG_LOAD_FILE);
		}
		else if(item.getItemId() == R.id.process) {
			if(this.mChosenFile!= null && this.mChosenFile.length()>3) {
				processImage();
			}
			else {
				Toast toast = Toast.makeText(this,
						"Please select a file first", Toast.LENGTH_LONG);
				toast.show();
			}
		}
		else if(item.getItemId() == R.id.settings) {
			startActivity(new Intent(this, Preferences.class));
			loadDigitArea();
			drawView.setDigitArea(digitArea);
		}
		else if(item.getItemId() == R.id.selectarea) {
			drawView.setAreaSelectable(!drawView.isAreaSelectable());
		}

		return true;
	}
    
    
    private void loadFileList() {
        try {
        	if(mPath.exists()) {
                FilenameFilter filter = new FilenameFilter() {
                    public boolean accept(File dir, String filename) {
                        File sel = new File(dir, filename);
                        return filename.contains(FTYPE) || sel.isDirectory();
                    }
                };
                mFileList = mPath.list(filter);
            }
            else {
                mFileList= new String[0];
            }
        }
        catch(SecurityException e) {
            Log.e(TAG, "unable to read the sd card " + e.toString());
        }
        
    }

    protected Dialog onCreateDialog(int id) {
    	loadFileList();
        AlertDialog dialog = null;
        AlertDialog.Builder builder = new Builder(this);

        switch(id) {
            case DIALOG_LOAD_FILE:
                builder.setTitle("Choose your file");
                if(mFileList == null) {
                    Log.e(TAG, "Showing file picker before loading the file list");
                    dialog = builder.create();
                    return dialog;
                }
                builder.setItems(mFileList, this);
                break;
        }
        dialog = builder.show();
        return dialog;
    }

    public void onClick(DialogInterface dialog, int which) {
        mChosenFile = mFileList[which]; 
        loadImage(mPath + "/" + mChosenFile);
        TextView txtCoord = (TextView) this.findViewById(R.id.coords);
        txtCoord.setVisibility(TextView.INVISIBLE);
        drawView = new DrawView(this, pixels, digitArea, txtCoord);
        drawView.setBackgroundColor(Color.BLACK);
        imageLayout.addView(drawView);
    }
    
    
    
    /**
	 * load the Luminance from file
	 * 
	 * @param fileName
	 * @throws Exception
	 */
	public void loadImage(String fileName)  {
			
		String ht = sharedPrefs.getString("height", "240");
		imgHeight = Integer.parseInt(ht);
		String wd = sharedPrefs.getString("width", "320");
		imgWidth = Integer.parseInt(wd);
		
		int len = imgWidth * imgHeight;
		int[] lum = new int[len];
		try {
			DataInputStream dis = new DataInputStream(
					new FileInputStream(fileName));
						
			for (int i = 0; i < (len); i++) {
				lum[i] = dis.readByte();
			}		
									
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
				
		
		pixels = new int[imgWidth][imgHeight];
		int lCt=0;
		
			for (int y = 0; y < pixels[0].length; y++) {
				for (int x = 0; x < pixels.length; x++){
				int c = lum[lCt++];
				pixels[x][y] = c;					
			}
		}
			//dumpPixels(0,0, pixels.length, pixels[0].length);
	}
    
    private void processImage() {
    	LcdDigitOcr ocr = loadOcr();
    	    	
    	//dumpPixels(0,0, pixels.length, pixels[0].length);
    	int[][] procPx = ocr.brightnessToMonochrome(pixels);
		drawView.setProcessedPixels(procPx);
		
		drawView.setTracedLines( ocr.traceSegments() );
		imageLayout.invalidate();
		
		String value = ocr.readDigits();
		Toast t = Toast.makeText(super.getApplicationContext(), value, Toast.LENGTH_LONG);
		t.show();
    }
    
    private void dumpPixels(int xmin, int ymin, int xmax, int ymax) {
		
		for(int y=ymin; y<ymax; y++) {
			for(int x=xmin; x<xmax; x++) {	
				if(pixels[x][y]>140) System.out.print(1);
				else System.out.print(0);
			}
			System.out.println();
		}
	}
    
    /**
	 * load the OCR 
	 */
	private LcdDigitOcr loadOcr() {
		
		//expecting "x1,y1,x2,y2"
		
		
		String sThreshold = sharedPrefs.getString("bw_threshold", "80");
		int threshold = Integer.parseInt(sThreshold);

		return new LcdDigitOcr(digitArea, threshold);
		
	}
}